import re
from inspect import isclass
from typing import Tuple, Optional
from dataclasses import dataclass

from hogwarts.magic_urls._base import (
    import_views,
    get_path_name,
    get_path_url,
    view_is_detail,
    has_path_decorator,
    get_decorator_path_name,
    get_decorator_path_url
)


@dataclass
class UtilityPath:
    path: str
    view_name: str


def gen_urls_py(views_module, urls_path: str, app_name, force_app_name: bool):
    imports = gen_url_imports(import_views(views_module), "views")
    if not force_app_name:
        app_name = read_app_name_from_urls_py(urls_path) or app_name

    urlpatterns = gen_urlpatterns(views_module, app_name)

    with open(f"{urls_path}", 'w') as file:
        file.write(imports + f'\n\n\napp_name = "{app_name}"' + urlpatterns)


def merge_urls_py(views_module, urls_path, app_name, force_app_name: bool):
    file = open(urls_path, "r")
    code = file.read()

    imports, urlpatterns = separate_imports_and_urlpatterns(code)

    if not force_app_name:
        app_name = get_app_name(code) or app_name

    views = import_views(views_module)
    paths: list[UtilityPath] = []

    for view in views:
        path = gen_path(view, app_name)
        view_name = view.__name__
        paths.append(UtilityPath(path, view_name))

    for view in views:
        if view.__name__ not in imports:
            imports = append_view_into_imports(imports, view)

    for path in paths:
        if path.view_name not in urlpatterns:
            urlpatterns = append_path_into_urlpatterns(urlpatterns, path.path)

    with open(urls_path, 'w') as file:
        file.write(imports + f'\n\napp_name = "{app_name}"\n' + urlpatterns)


def gen_url_imports(views: list[object], views_file_name: str):
    view_names = ", ".join((view.__name__ for view in views))

    return f"from django.urls import path\n\nfrom .{views_file_name} import {view_names}"


def gen_urlpatterns(views_module, app_name: str):
    views = import_views(views_module)
    urlpatterns = []

    for view in views:
        urlpatterns.append(gen_path(view, app_name))

    paths_string = ",\n    ".join(urlpatterns)

    result = f"""
urlpatterns = [
    {paths_string.strip()}
]    
    """

    return result


def gen_path(view, app_name) -> str:
    if has_path_decorator(view):
        path_name = get_decorator_path_name(view)
        path_url = get_decorator_path_url(view)

        print(
            f"info: You have provided @custom_path decorator in {view.__name__}"
            f"don't forget to remove because it is not in use anymore"
        )

    else:
        path_name = get_path_name(view, app_name)
        path_url = get_path_url(
            path_name,
            detail=view_is_detail(view) if isclass(view) else False
        )

    if isclass(view):
        view_function = f"{view.__name__}.as_view()"
    else:
        view_function = f"{view.__name__}"

    return f'path("{path_url}", {view_function}, name="{path_name}")'


def separate_imports_and_urlpatterns(code: str) -> Tuple[str, str]:
    imports = ""
    urlpatterns = ""

    lines = code.splitlines()

    for i in range(len(lines)):
        if lines[i].startswith("from"):
            imports += lines[i] + "\n"

            if lines[i].endswith("("):
                x = i + 1
                while not lines[x].endswith(")"):
                    imports += lines[x] + "\n"
                    x += 1
                imports += lines[x] + "\n"

        if lines[i].startswith("urlpatterns"):
            for x in range(i, len(lines)):
                urlpatterns += lines[x] + "\n"
            break

    return imports, urlpatterns


def append_view_into_imports(imports: str, view: str):
    lines = imports.splitlines()

    for i in range(len(lines)):
        if lines[i].startswith("from .views"):
            if lines[i].endswith("("):
                x = i + 1
                while not lines[x].endswith(")"):
                    x += 1

                if lines[x].strip() == ")":
                    lines[x - 1] += ","
                    lines[x] = lines[x].replace(")", f"    {view}\n)")
                else:
                    lines[x] = lines[x].replace(")", f", {view})")

                break

            tokens = lines[i].split(" ")
            if view not in tokens:
                lines[i] += ", " + view

            break

    return "\n".join(lines)


def append_path_into_urlpatterns(urlpatterns: str, path: str):
    lines = urlpatterns.splitlines()
    lines = [line.rstrip() for line in lines]
    bracket = None

    for i, line in enumerate(lines):
        if "]" in line:
            bracket = i
            break

    if bracket:
        if not lines[bracket - 1].endswith(","):
            lines[bracket - 1] = lines[bracket - 1] + ","

        lines[bracket] = "    " + path
        lines.append("]")

    return "\n".join(lines)


def get_app_name(code: str) -> Optional[str]:
    lines = code.splitlines()

    for line in lines:
        if line.startswith("app_name"):
            pattern = r'"([^"]*)"'
            match = re.search(pattern, line)
            return match.group(1)

    return None


def urlpatterns_is_empty(code):
    return bool(re.search(r'urlpatterns\s*=\s*\[\s*\]', code))


def read_app_name_from_urls_py(urls_path: str):
    file = open(urls_path, "r")
    code = file.read()

    imports, urlpatterns = separate_imports_and_urlpatterns(code)
    return get_app_name(code)



from django.db import models


# Create your models here.
class Example(models.Model):
    message = models.CharField(max_length=255)


class Article(models.Model):
    title = models.CharField(max_length=255)
    description = models.TextField()
    created_at = models.DateTimeField(auto_now_add=True)
    beta = models.BooleanField(default=False)

    


# Generated by Django 4.2.2 on 2023-06-21 16:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("hogwarts", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Article",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=255)),
                ("description", models.TextField()),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("beta", models.BooleanField(default=False)),
            ],
        ),
    ]



from dataclasses import dataclass


@dataclass
class Path:
    path_url: str
    view: str
    path_name: str


def extract_paths_from_urls_code(code: str):
    lines = code.strip().split("\n")
    extracted_list: list[Path] = []

    for line in lines:
        if "path(" in line:
            parts = line.split('"')
            first_arg = parts[1]
            second_arg = parts[2].split(',')[1].strip()
            if second_arg.endswith(".as_view()"):
                second_arg = second_arg.split(".")[0]

            third_arg = parts[3]
            extracted_list.append(Path(first_arg, second_arg, third_arg))

    return extracted_list


def append_path_to_urls_code(code: str, path: str):
    closing_bracket_index = code.rfind(']')

    if closing_bracket_index == -1:
        raise ValueError(f"code {code} does not have closing ']' symbol")

    return code[:closing_bracket_index] + f"    {path}\n]"



from ..magic_urls.genurls import gen_urlpatterns, gen_path, gen_url_imports
from ..magic_urls import custom_path

from .. import _test_views


def test_it_generates_path():
    result = gen_path(_test_views.MyListView, "my")
    expected = 'path("", MyListView.as_view(), name="list")'

    assert result == expected


def test_it_generates_path_for_function():
    result = gen_path(_test_views.confirm_post_view, "none")
    expected = 'path("confirm-post/", confirm_post_view, name="confirm_post")'

    assert result == expected


def test_it_extracts_metadata():
    @custom_path("green", "green-hello/")
    class RedView:
        pass

    result = gen_path(RedView, "none")
    expected = 'path("green-hello/", RedView.as_view(), name="green")'

    assert result == expected


def test_it_generates_urls():
    result = gen_urlpatterns(_test_views, "my")

    expected = """
urlpatterns = [
    path("form/", MyFormView.as_view(), name="form"),
    path("", MyListView.as_view(), name="list"),
    path("confirm-post/", confirm_post_view, name="confirm_post"),
    path("get/", get_view, name="get"),
    path("post/", post_view, name="post")
]    
    """

    assert result == expected


def test_it_generates_imports():
    result = gen_url_imports([
        _test_views.MyListView,
        _test_views.MyFormView,
        _test_views.get_view
    ], "_test_views")

    expected = """from ._test_views import MyListView, MyFormView, get_view"""

    assert result == expected



import importlib

from django.core.management.base import CommandError
from django.apps import apps


def get_app_config(app_name: str):
    try:
        return apps.get_app_config(app_name)
    except LookupError:
        raise CommandError(f"App '{app_name}' does not exist.")


def get_views_module(app_name: str):
    try:
        # Import the views.py file dynamically
        return importlib.import_module(f"{app_name}.views")
    except ModuleNotFoundError:
        raise CommandError(f"Views not found for app '{app_name}'.")



from hogwarts.codegen import ViewGenerator, insert_code

from ..models import Article
from ..utils import code_strip

generator = ViewGenerator(Article)


def test_it_generates_detail_view():
    code = generator.gen_detail_view()
    expected_code = """
    class ArticleDetailView(DetailView):
        model = Article
        context_object_name = "article"
        template_name = "articles/article_detail.html"
    """

    assert code_strip(code) == code_strip(expected_code)


def test_it_generates_list_view():
    code = generator.gen_list_view()
    expected_code = """
    class ArticleListView(ListView):
        model = Article
        context_object_name = "articles"
        template_name = "articles/article_list.html"
    """

    assert code_strip(code) == code_strip(expected_code)


def test_it_generated_create_view():
    code = generator.gen_create_view()
    expected_code = """
    class ArticleCreateView(CreateView):
        model = Article
        fields = ['id', 'title', 'description', 'created_at', 'beta']
        template_name = "articles/article_create.html"
    """

    assert code_strip(code) == code_strip(expected_code)


def test_it_generated_update_view():
    code = generator.gen_update_view()
    expected_code = """
    class ArticleUpdateView(UpdateView):
        model = Article
        fields = ['id', 'title', 'description', 'created_at', 'beta']
        template_name = "articles/article_update.html"
    """

    assert code_strip(code) == code_strip(expected_code)


def test_code_gen_imports():
    gen = ViewGenerator(Article)
    gen.gen_update_view()
    gen.gen_detail_view()

    imports = ["UpdateView", "DetailView", "Article"]
    imports_code = """
    from django.views.generic import UpdateView, DetailView
    from .models import Article
    """

    assert set(gen.imports) == set(imports)
    assert code_strip(gen.get_imports_code()) == code_strip(imports_code)


def test_it_inserts_code():
    gen = ViewGenerator(Article)
    create_code = gen.gen_create_view()
    detail_code = gen.gen_detail_view()

    new_code = insert_code([create_code, detail_code], gen.get_imports_code())

    expected_code = """
    from django.views.generic import CreateView, DetailView
    from .models import Article
    
    class ArticleCreateView(CreateView):
        model = Article
        fields = ['id', 'title', 'description', 'created_at', 'beta']
        template_name = "articles/article_create.html"
    
    class ArticleDetailView(DetailView):
        model = Article
        context_object_name = "article"
        template_name = "articles/article_detail.html"
    """

    print(new_code)
    print("======")
    print(code_strip(expected_code))

    assert code_strip(new_code) == code_strip(expected_code)



from typing import Type
from dataclasses import dataclass

from django.db import models

from .utils import to_plural, code_strip, remove_empty_lines

@dataclass
class ClassView:
    imports: set[str]
    name: str
    code: str


class ViewGenerator:
    def __init__(self, model: Type[models.Model]):
        self.model = model
        self.model_name = model.__name__
        self.model_name_lower = self.model_name.lower()

        self.fields = model._meta.fields
        self.field_names = [field.name for field in self.fields]
        self.imports = [self.model_name]
        self.class_names = []

    def get_imports_code(self):
        imports = self.imports[1:]
        return f"""
        from django.views.generic import {", ".join(imports)}
        from .models import {self.model_name}
        """

    def gen_detail_view(self):
        self.imports.append("DetailView")
        self.add_class(f"{self.model_name}DetailView")
        return self.base_view("detail", False, True, True)

    def gen_list_view(self):
        self.imports.append("ListView")
        self.add_class(f"{self.model_name}ListView")
        return self.base_view("list", False, True)


    def gen_create_view(self):
        self.imports.append("CreateView")
        self.add_class(f"{self.model_name}CreateView")
        return self.base_view("create", True, False)

    def gen_update_view(self):
        self.imports.append("UpdateView")
        self.add_class(f"{self.model_name}UpdateView")
        return self.base_view("update", True, False)

    def base_view(self, action: str, fields: bool, context: bool, detail: bool = False):
        name = self.model_name_lower
        action_view = f"{action.capitalize()}View"
        object_name = name if detail else to_plural(name)
        template_name = f"{to_plural(name)}/{name}_{action.lower()}.html"

        result = f"""
        class {self.model_name}{action_view}({action_view}):
            model = {self.model_name}
            {f'fields = {str(self.field_names)}' if fields else ''}
            {f'context_object_name = "{object_name}"' if context else ''}
            template_name = "{template_name}"
        """

        return remove_empty_lines(result) + "\n"

    def add_class(self, class_name: str):
        if class_name not in self.class_names:
            self.class_names.append(class_name)


def insert_code(code_blocks: list[str], imports: str):
    origin_code = code_strip(imports)

    for code in code_blocks:
        origin_code += f"\n{code_strip(code)}"

    return origin_code


def generate_views(model: Type[models.Model]):
    gen = ViewGenerator(model)

    detail = gen.gen_detail_view()
    _list = gen.gen_list_view()
    create = gen.gen_create_view()
    update = gen.gen_update_view()

    return insert_code([detail, _list, create, update], gen.get_imports_code())



from django.core.management.base import BaseCommand, CommandError
from django.apps import apps

from hogwarts.codegen import generate_views


class Command(BaseCommand):
    help = "Code generation command"

    def add_arguments(self, parser):
        parser.add_argument("app", type=str)

    def handle(self, *args, **options):
        app_name: str = options["app"]
        model_name: str = options["model"]

        app_names = [_app.name for _app in apps.get_app_configs()]

        if app_name not in app_names:
            raise CommandError(f"Provided app '{app_name}' does not exist")

        app_config = apps.get_app_config(app_name)

        if model_name.lower() not in app_config.models.keys():
            raise CommandError(f"Provided model '{model_name}' does not exist in app '{app_name}'")

        model = app_config.models.get(model_name.lower())
        code = generate_views(model)

        path = f'{app_config.path}\\generated_views.py'
        with open(path, 'w') as file:
            file.write(code)

        self.stdout.write(
            self.style.SUCCESS(f"Generated CRUD views in {path}")
        )



from django.core.management.base import BaseCommand
from rich.console import Console
from rich.syntax import Syntax

from hogwarts.magic_urls.genurls import gen_urls_py, merge_urls_py, urlpatterns_is_empty
from .base import get_app_config, get_views_module


console = Console()


class Command(BaseCommand):
    help = "urlpatterns generation from views.py"

    def add_arguments(self, parser):
        parser.add_argument("app_name", type=str)
        parser.add_argument(
            "--merge", "-m",
            action="store_true",
            help="add urls to existing urlpatterns without overriding whole file"
        )

        parser.add_argument(
            "--force-app-name", "-fan",
            action="store_true",
            help="use given app name rather than app_name in urls.py"
        )

    def handle(self, *args, **options):
        app_name: str = options["app_name"]
        merge: bool = options["merge"]
        force_new_app_name: bool = options["force_app_name"]

        views_module = get_views_module(app_name)
        app_config = get_app_config(app_name)
        urls_path = f'{app_config.path}\\urls.py'

        if merge:
            merge_urls_py(views_module, urls_path, app_name, force_new_app_name)
            console.print("new paths merged to urlpatterns ✅", style="green")
        else:
            code = open(urls_path, "r").read()
            if not urlpatterns_is_empty(code):
                # alert user that urls is not empty
                console.print(
                    "\nLooks like you have some paths in urlpatterns. 🚧\n"
                    f"{urls_path}:",
                    style="bold yellow"
                )
                print("===================")
                Console().print(Syntax(code, "python"))
                print("===================")

                # get user input until correct
                while True:
                    print("Do you want fully override urls or merge?")
                    response = input("write (o)-override, (m)-merge or (c) to cancel: ")
                    if response not in ["m", "o", "c"]:
                        print("wrong command!")
                        continue

                    if response == "m":
                        merge_urls_py(views_module, urls_path, app_name, force_new_app_name)
                        console.print("new paths merged to urlpatterns ✅", style="green")

                    elif response == "o":
                        gen_urls_py(views_module, urls_path, app_name, force_new_app_name)
                        console.print("urlpatterns have been generated ✅", style="green")

                    elif response == "c":
                        print("canceled")
                    return
            else:
                gen_urls_py(views_module, urls_path, app_name, force_new_app_name)
                console.print("urlpatterns have been generated ✅", style="green")




# Generated by Django 3.2.3 on 2023-04-15 03:04

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Example',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('message', models.CharField(max_length=255)),
            ],
        ),
    ]



from inspect import isclass
from typing import Optional

from django.urls import path

from ._base import (
    import_views,
    get_path_name,
    get_path_url,
    view_is_detail,
    has_path_decorator,
    get_decorator_path_name,
    get_decorator_path_url
)


def auto_urls(views_module, app_name: str):
    views = import_views(views_module)
    urlpatterns = []

    for view in views:
        urlpatterns.append(get_path(view, app_name))

    return urlpatterns


def get_path(view, app_name: Optional[str] = None):
    if has_path_decorator(view):
        path_name = get_decorator_path_name(view)
        path_url = get_decorator_path_url(view)

    else:
        path_name = get_path_name(view, app_name)
        path_url = get_path_url(path_name, detail=view_is_detail(view))

    if isclass(view):
        return path(path_url, view.as_view(), name=path_name)
    else:
        return path(path_url, view, name=path_name)



from hogwarts.magic_urls.genurls import gen_path
from hogwarts.magic_urls.utils import extract_paths_from_urls_code, append_path_to_urls_code


def test_it_extracts_paths():
    payload = """
urlpatterns = [
    path("", PostListView.as_view(), name="list"),
    path("create/", PostCreateView.as_view(), name="create"),
    path("comment-confirm/", comment_confirm_view, name="comment_confirm")
]       
    """

    result = extract_paths_from_urls_code(payload)

    assert result[0].path_name == "list"
    assert result[0].view == "PostListView"
    assert result[0].path_url == ""


def test_it_appends_paths():
    code = """
urlpatterns = [
    path("", PostListView.as_view(), name="list"),
    path("create/", PostCreateView.as_view(), name="create"),
]    
    """

    expected = """
urlpatterns = [
    path("", PostListView.as_view(), name="list"),
    path("create/", PostCreateView.as_view(), name="create"),
    path("<int:pk>/", PostDetailView.as_view(), name="detail")
]"""

    class PostDetailView:
        pass

    path = gen_path(PostDetailView, "post")
    result = append_path_to_urls_code(code, path)

    assert result == expected



import ast


def to_plural(string: str):
    return string if string.endswith("s") else f"{string}s"


def get_imports_list(path: str) -> list[str]:
    with open(path, "r") as file:
        tree = ast.parse(file.read())

    imports = []

    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                imports.append(alias.name)
        elif isinstance(node, ast.ImportFrom):
            for alias in node.names:
                imports.append(alias.name)

    return imports


def get_classes_list(path: str) -> list[str]:
    with open(path, "r") as file:
        tree = ast.parse(file.read())

    classes = []

    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef):
            classes.append(node.name)

    return classes

def code_strip(code):
    if code[0] == '\n':
        code = code[1:]

    spaces = 0
    code = code.split("\n")

    for i in code[0]:
        if i == " ":
            spaces += 1
        else:
            break

    for i in range(len(code)):
        code[i] = code[i][spaces:].rstrip()

    return "\n".join(code)


def remove_empty_lines(text):
    lines = text.splitlines()
    non_empty_lines = [line for line in lines if line.strip()]
    return '\n'.join(non_empty_lines)




# test import


class MyFormView:
    pass


class MyListView:
    pass


class SomeAnotherClass:
    pass


class AnotherOne(SomeAnotherClass):
    pass


def post_view(request):
    pass


def confirm_post_view(request):
    pass


def get_view(request):
    pass


def just_function():
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hogwarts',
 'hogwarts.magic_urls',
 'hogwarts.management',
 'hogwarts.management.commands',
 'hogwarts.migrations',
 'hogwarts.tests']

package_data = \
{'': ['*']}

install_requires = \
['Django>=4.2.2,<5.0.0', 'rich>=13.5.2,<14.0.0', 'ruff>=0.0.280,<0.0.281']

setup_kwargs = {
    'name': 'django-hogwarts',
    'version': '0.1.1',
    'description': 'Django utilities for codegen and DX improvement',
    'long_description': '<h1 align="center">Django hogwarts</h1>\n<h4 align="center">bring magic to you development ✨</h4>\n\nSet of utilities and CLI tools to speed up django development\nand do repetitive stuff\n\n# THIS REPO IS UNDER CONSTRUCTION 🚧\n### first beta will be till september 2023\n',
    'author': 'adiletto64',
    'author_email': 'adiletdj19@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/adiletto64/django_hogwarts',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import inspect
import re
from dataclasses import dataclass
from enum import Enum
from inspect import isclass
from typing import Optional, Type

from django.views import View
from django.views.generic import DeleteView, DetailView, UpdateView


detail_names = [
    "Detail",
    "Update",
    "Delete"
]


BASE_CLASS_NAMES = [
    'View',
    'ListView',
    'CreateView',
    'FormView',
    'DetailView',
    'DeleteView',
    'UpdateView',
]


class ViewType(Enum):
    FUNCTION = 1
    CLASS = 2


@dataclass
class Path:
    path_name: str
    detail: bool
    path_urls: Optional[str]


def get_path_name(view, app_name: Optional[str] = None):
    name: str = view.__name__
    view_type = ViewType.CLASS if isclass(view) else ViewType.FUNCTION

    if app_name:
        if view_type == ViewType.CLASS:
            app_name = app_name.capitalize()

        if app_name in name:
            name = name.replace(app_name, '')

        elif app_name.endswith('s') and app_name[:-1] in name:
            name = name.replace(app_name[:-1], '')

    name = name.replace('View', '').replace('view', '')
    name = name.strip("_")

    return camel_to_snake(name)


def get_path_url(path_name: str, detail=False):
    if not detail and path_name == "list":
        return ""

    path_url = path_name.replace("_", "-") + "/"
    if detail:
        if path_name == 'detail':
            path_url = "<int:pk>/"
        else:
            path_url = "<int:pk>/" + path_url

    return path_url


def view_is_detail(view_class: Type[View]):
    for cls in [DeleteView, DetailView, UpdateView]:
        if issubclass(view_class, cls):
            return True

    return any(ending in view_class.__name__ for ending in detail_names)


def camel_to_snake(camel_case_string):
    snake_case_string = re.sub(r'(?<=[a-z])([A-Z])', r'_\1', camel_case_string)
    snake_case_string = snake_case_string.lower()
    return snake_case_string


def import_views(views_module):
    members = inspect.getmembers(views_module, predicate=is_view)
    return [t[1] for t in members]


def is_view(obj):
    try:
        name = obj.__name__
    except AttributeError:
        return False

    ends_with_view = name.lower().endswith("view")
    not_base_class = name not in BASE_CLASS_NAMES

    return ends_with_view and not_base_class


def custom_path(
        path_name: str,
        path_url: Optional[str] = None,
        detail: bool = False
):
    def wrapper(obj):
        obj.auto_url_path = Path(path_name, detail, path_url)
        return obj

    return wrapper


def has_path_decorator(view):
    return hasattr(view, "auto_url_path")


def get_decorator_path_name(view):
    path: Path = getattr(view, "auto_url_path")
    return path.path_name


def get_decorator_path_url(view):
    path: Path = getattr(view, "auto_url_path")
    path_name = get_decorator_path_name(view)
    return path.path_urls or get_path_url(path_name, detail=path.detail)
